#! /usr/bin/env python
__author__ = "John Hover, Jose Caballero"
__copyright__ = "2017 John Hover"
__credits__ = []
__license__ = "GPL"
__version__ = "0.9.1"
__maintainer__ = "John Hover"
__email__ = "jhover@bnl.gov"
__status__ = "Production"

# utilites for devel, debug
# consume python config files, interpolate, emit human-READABLE
# consume file and urlencode

class UtilCLI(object):
    def __init__(self):
        pass


if __name__ == "__main__":
    pass


